from flask_admin import Admin, AdminIndexView, expose
from flask_admin.contrib.sqla import ModelView
from flask_admin.model import InlineFormAdmin
from flask_login import current_user
from flask import redirect, url_for, request

from kirby.models import ConfigKey, NotificationEmail


def is_authenticated(user):
    return user.is_authenticated and user.is_active and not user.is_anonymous


class AuthenticatedModelView(ModelView):
    def is_accessible(self):
        return is_authenticated(current_user)

    def inaccessible_callback(self, name, **kwargs):
        return redirect(url_for("security.login", next=request.url))


class UserView(AuthenticatedModelView):
    form_excluded_columns = ("password", "provider")
    column_exclude_list = ("password",)

    def on_form_prefill(self, form, id):
        form.username.render_kw = {"readonly": True}


class ConfigKeyView(AuthenticatedModelView):
    form_excluded_columns = ("job", "context", "scope")
    column_list = ("job_name", "environment_name", "scope", "name", "value")
    column_searchable_list = (
        "scope",
        "name",
        "job.name",
        "context.job.name",
        "context.environment.name",
    )

    def on_form_prefill(self, form, id):
        form.name.render_kw = {"readonly": True}


class ContextConfigKeyInlineModel(InlineFormAdmin):
    form_excluded_columns = ("job", "scope")


class ContextView(AuthenticatedModelView):
    inline_models = [ContextConfigKeyInlineModel(ConfigKey)]


class JobConfigKeyInlineModel(InlineFormAdmin):
    form_excluded_columns = ("context", "scope")


class JobView(AuthenticatedModelView):
    inline_models = [JobConfigKeyInlineModel(ConfigKey)]


class NotificationGroupView(AuthenticatedModelView):
    form_excluded_columns = ("notifications",)
    inline_models = [InlineFormAdmin(NotificationEmail)]


class ScriptView(AuthenticatedModelView):
    form_excluded_columns = (
        "sources",
        "destinations",
        "first_seen",
        "last_seen",
    )


class KirbyAdminIndexView(AdminIndexView):
    @expose("/")
    def index(self):
        if not is_authenticated(current_user):
            return redirect(url_for("security.login", next=request.url))
        return super().index()


from ..models import (
    db,
    Environment,
    Job,
    Schedule,
    Context,
    ConfigKey,
    Suspension,
    NotificationGroup,
    Notification,
    Script,
    Topic,
)
from ..models.security import User

admin = Admin(
    name="Kirby",
    url="/admin",
    template_mode="bootstrap3",
    base_template="kirby_master.html",
    index_view=KirbyAdminIndexView(url="/admin"),
)

models = {
    "Scheduling": [Schedule, Suspension],
    "Notifications": [Notification],
    "Jobs": [Environment],
    "Documentation": [Topic],
}


for category, models in models.items():
    for model in models:
        admin.add_view(
            AuthenticatedModelView(model, db.session, category=category)
        )


admin.add_view(UserView(User, db.session, category="Users"))
admin.add_view(
    ConfigKeyView(ConfigKey, db.session, category="Jobs", name="Configuration")
)
admin.add_view(ContextView(Context, db.session, category="Jobs"))
admin.add_view(JobView(Job, db.session, category="Jobs"))
admin.add_view(
    NotificationGroupView(
        NotificationGroup, db.session, category="Notifications"
    )
)
admin.add_view(ScriptView(Script, db.session, category="Jobs"))



import os
from smart_getenv import getenv
import pickle
import codecs

KIRBY_ENV_SIGNATURE = "__KIRBY_ENV_SIGNATURE"


class ContextManager:
    def __init__(self, config):
        self.config = config
        os.environ[KIRBY_ENV_SIGNATURE] = codecs.encode(
            pickle.dumps(self.config), "base64"
        ).decode()


def get_signature():
    return pickle.loads(
        codecs.decode(os.environ[KIRBY_ENV_SIGNATURE].encode(), "base64")
    )


class Context:
    def __getattr__(self, item):
        signature = get_signature().get(item, {"type": str})
        return getenv(item, **signature)

    def __repr__(self):
        all_signatures = get_signature()
        return "; ".join(
            [
                f"{var}={repr(getenv(var, **signature))}"
                for var, signature in all_signatures.items()
            ]
        )

    def __getitem__(self, item):
        return self.__getattr__(item)


ctx = Context()



from kirby.models.security import user_datastore
from kirby.models import db


def test_create_user(webapp):
    user = user_datastore.create_user(
        username="admin", email="admin@local.local", password="password"
    )
    db.session.commit()

    assert user.is_local
    assert not user.is_admin


def test_promote_user_to_admin(webapp):

    user = user_datastore.create_user(
        username="admin", email="admin@local.local", password="password"
    )
    role = user_datastore.find_role("admin")
    assert role

    db.session.commit()

    assert not user.is_admin

    user_datastore.add_role_to_user(role=role, user=user)

    db.session.commit()

    assert user.is_admin



from datetime import datetime
from freezegun import freeze_time

from kirby.models import db, Schedule, Script
from kirby.web.endpoints import should_run

from tests.conftest import API_ROOT


def test_it_compares_correctly_schedule_and_time():
    date = datetime(2019, 5, 20, 12, 0)
    assert should_run(Schedule(name="Every minute"), date)
    assert should_run(
        Schedule(
            name="At 12:00 on day-of-month 20", minute="0", hour="12", day="20"
        ),
        date,
    )
    assert not should_run(
        Schedule(name="Every minute past hour 13", hour="13"), date
    )


def test_it_get_the_current_schedule(session, db_scripts_registered):
    result = session.get("/".join([API_ROOT, "schedule"]))
    result_json = result.json()

    assert result.status_code == 200
    assert result_json["jobs"][0]["environment"] == "test_env"


def get_topic_id(session, name):
    response = session.get(
        "/".join([API_ROOT, "topic"]), params={"name": name}
    )
    assert response.status_code == 200
    return response.json()["id"]


def test_if_get_topic_id(session, db_topics):
    for i, topic in enumerate(db_topics, start=1):
        assert get_topic_id(session, topic.name) == i


@freeze_time("2019-05-20 15:34")
def test_it_register_a_script(session, db_scripts_not_registered, db_topics):
    id_source = get_topic_id(session, "orders")
    id_destination = get_topic_id(session, "asset_management")

    result = session.patch(
        "/".join([API_ROOT, "registration"]),
        params={
            "script_id": 1,
            "source_id": id_source,
            "destination_id": id_destination,
        },
    )
    assert result.status_code == 200

    script_registered = (
        db.session.query(Script)
        .filter_by(package_name="orders_retriever")
        .one()
    )
    assert script_registered.sources[0].id == id_source
    assert script_registered.destinations[0].id == id_destination


def test_schedule_contains_configuration(session, db_scripts_registered):
    script = (
        db.session.query(Script)
        .filter_by(package_name="orders_retriever")
        .one()
    )

    context = script.context
    context.set_config(API_USERNAME="test-user", API_PORT=2000)

    result = session.get("/".join([API_ROOT, "schedule"]))
    result_json = result.json()

    assert result.status_code == 200
    assert result_json["jobs"][0]["variables"] == [
        {"key": "API_USERNAME", "value": "test-user"},
        {"key": "API_PORT", "value": "2000"},
    ]



import requests
import logging

from .context import ContextManager, ctx

logger = logging.getLogger(__name__)


class ServerError(BaseException):
    pass


class ClientError(BaseException):
    pass


class Kirby:
    def __init__(self, env_signature, session=None):
        ContextManager(env_signature)
        self.ctx = ctx
        self._session = session or requests.session()

        self._register()

    def get_topic_id(self, topic_name):
        result = self._session.get(
            "/".join([self.ctx.KIRBY_WEB_SERVER, "topic"]),
            params={"name": topic_name},
        )
        if result.status_code != 200:
            if result.status_code == 500:
                raise ServerError("There is an issue with the web server.")
            else:
                raise ClientError(
                    f"There is no id from the name {topic_name}. "
                    f"Verify that the topic has been registered."
                )
        return result.json()["id"]

    def _register(self, source_id=None, destination_id=None):
        result = self._session.patch(
            "/".join([self.ctx.KIRBY_WEB_SERVER, "registration"]),
            params={
                "script_id": self.ctx.ID,
                "source_id": source_id,
                "destination_id": destination_id,
            },
        )

        if result.status_code != 200:
            if result.status_code == 500:
                raise ServerError("There is an issue with the web server.")
            else:
                raise ClientError("There is an error with the id(s) given.")

    def add_source(self, source):
        self._register(source_id=self.get_topic_id(source.name))

    def add_destination(self, destination):
        self._register(destination_id=self.get_topic_id(destination.name))



from pytest import fixture
import os
from requests_flask_adapter import Session


from kirby.web import app_maker
from kirby.models import (
    db,
    Environment,
    JobType,
    Job,
    Context,
    Schedule,
    NotificationGroup,
    NotificationEmail,
    Notification,
    Script,
    Topic,
)

API_ROOT = "http://some-test-server.somewhere"


@fixture(scope="function")
def webapp():
    app = app_maker(
        config={
            "TESTING": True,
            "SQLALCHEMY_DATABASE_URI": "sqlite://",
            "SQLALCHEMY_TRACK_MODIFICATIONS": False,
        }
    )
    with app.app_context():
        app.try_trigger_before_first_request_functions()
        yield app


@fixture
def data_dir():
    current_dir = os.path.dirname(__file__)
    data_dir = os.path.join(current_dir, "data")
    return data_dir


@fixture
def session(webapp):
    Session.register(API_ROOT, webapp)
    return Session()


@fixture
def db_env(session):
    env = Environment(name="test_env")
    db.session.add(env)
    db.session.commit()
    return env


@fixture
def db_notification_groups(session):
    admin = NotificationGroup(
        name="Admin",
        emails=[NotificationEmail(email="admin@some-test-server.somewhere")],
    )
    producers = NotificationGroup(
        name="Producers",
        emails=[
            NotificationEmail(email="producer1@some-test-server.somewhere"),
            NotificationEmail(email="producer2@some-test-server.somewhere"),
        ],
    )
    bakery = NotificationGroup(
        name="Bakery",
        emails=[NotificationEmail(email="bakery@some-test-server.somewhere")],
    )

    notification_groups = [admin, producers, bakery]
    db.session.add_all(notification_groups)
    db.session.commit()
    return notification_groups


@fixture
def db_jobs(db_notification_groups):
    [admin, producers, bakery] = db_notification_groups
    orders_job = Job(
        name="Fetch Orders",
        type=JobType.SCHEDULED,
        notifications=[
            Notification(on_retry=True, on_failure=True, groups=[admin]),
            Notification(
                on_retry=False, on_failure=True, groups=[producers, bakery]
            ),
        ],
    )
    cashregister_job = Job(
        name="Collect Register",
        type=JobType.TRIGGERED,
        notifications=[
            Notification(on_retry=True, on_failure=True, groups=[admin])
        ],
    )

    prepare_job = Job(
        name="Prepare Daily Orders",
        type=JobType.SCHEDULED,
        notifications=[
            Notification(on_retry=True, on_failure=True, groups=[admin])
        ],
    )
    abort_job = Job(
        name="Stop everything",
        type=JobType.SCHEDULED,
        notifications=[
            Notification(on_retry=True, on_failure=True, groups=[admin])
        ],
    )
    jobs = [orders_job, cashregister_job, prepare_job, abort_job]
    db.session.add_all(jobs)
    db.session.commit()
    return jobs


@fixture
def db_contexts(db_env, db_jobs):
    [orders_job, cashregister_job, prepare_job, abort_job] = db_jobs
    orders_context = Context(
        environment=db_env,
        job=orders_job,
        schedules=[Schedule(name="Every minute")],
    )
    cashregister_context = Context(environment=db_env, job=cashregister_job)
    prepare_context = Context(
        environment=db_env,
        job=prepare_job,
        schedules=[Schedule(name="Every day at 00:00", minute="0", hour="0")],
    )
    abort_context = Context(environment=db_env, job=abort_job)

    contexts = [
        orders_context,
        cashregister_context,
        prepare_context,
        abort_context,
    ]
    db.session.add_all(contexts)
    db.session.commit()
    return contexts


@fixture
def db_scripts_not_registered(db_contexts):
    [
        orders_context,
        cashregister_context,
        prepare_context,
        abort_context,
    ] = db_contexts
    orders_script = Script(
        package_name="orders_retriever",
        package_version="3.1.0",
        context=orders_context,
    )
    cashregister_script = Script(
        package_name="cashregister_retriever",
        package_version="2.0.4",
        context=cashregister_context,
    )
    prepare_script = Script(
        package_name="prepare_order_for_factory",
        package_version="2.2.1",
        context=prepare_context,
    )
    abort_script = Script(
        package_name="abort_all",
        package_version="0.0.1",
        context=abort_context,
    )

    scripts = [
        orders_script,
        cashregister_script,
        prepare_script,
        abort_script,
    ]
    db.session.add_all(scripts)
    db.session.commit()
    return scripts


@fixture
def db_topics(session):
    cashregister = Topic(name="cashregister")
    orders = Topic(name="orders")
    errors_log = Topic(name="errors_log")
    asset_management = Topic(name="asset_management")
    factory = Topic(name="factory")
    timeseries = Topic(name="timeseries")

    topics = [
        cashregister,
        orders,
        errors_log,
        asset_management,
        factory,
        timeseries,
    ]
    db.session.add_all(topics)
    db.session.commit()
    return topics


@fixture
def db_scripts_registered(db_scripts_not_registered, db_topics):
    [
        cashregister,
        orders,
        errors_log,
        asset_management,
        factory,
        timeseries,
    ] = db_topics
    [
        orders_script,
        cashregister_script,
        prepare_script,
        abort_script,
    ] = db_scripts_not_registered

    orders_script.add_source(orders)
    orders_script.add_destination(timeseries)
    orders_script.add_destination(asset_management)

    cashregister_script.add_source(cashregister)
    cashregister_script.add_destination(timeseries)
    cashregister_script.add_destination(asset_management)

    prepare_script.add_source(asset_management)
    prepare_script.add_destination(factory)

    abort_script.add_destination(errors_log)
    db.session.commit()


@fixture(scope="function")
def kafka_topic_factory():
    from smart_getenv import getenv
    from contextlib import contextmanager
    from kafka import KafkaAdminClient
    from kafka.admin import NewTopic
    from kafka.errors import UnknownTopicOrPartitionError
    import logging

    logger = logging.getLogger(__name__)

    bootstrap_servers = getenv(
        "KAFKA_BOOTSTRAP_SERVERS", type=list, separator=","
    )
    if bootstrap_servers:
        security_protocol = getenv("KAFKA_SECURITY_PROTOCOL")
        if security_protocol == "SSL":
            args = {
                "security_protocol": security_protocol,
                "ssl_cafile": getenv("KAFKA_SSL_CAFILE"),
                "ssl_certfile": getenv("KAFKA_SSL_CERTFILE"),
                "ssl_keyfile": getenv("KAFKA_SSL_KEYFILE"),
            }
        else:
            args = {}

        admin = KafkaAdminClient(bootstrap_servers=bootstrap_servers, **args)

        @contextmanager
        def create_kafka_topic(topic_name):
            try:
                admin.delete_topics([topic_name])
            except UnknownTopicOrPartitionError:
                pass

            admin.create_topics([NewTopic(topic_name, 1, 1)], timeout_ms=1500)
            yield

            admin.delete_topics([topic_name])
            admin.close()

        return create_kafka_topic

    else:
        logger.warning(
            f"There is no KAFKA_BOOTSTRAP_SERVERS. "
            f"Creation of kafka_topic skipped."
        )



import logging

from threading import Thread, Event

LEADER_KEY = "_KIRBY_LEADER"
LEADER_KEY_LEASE = 0.5  # seconds

logger = logging.getLogger(__name__)


class Timer(Thread):
    """Call a function after a specified number of seconds:

    t = Timer(30.0, f, args=[], kwargs={})
    t.start()
    t.cancel() # stop the timer's action if it's still waiting
    """

    def __init__(self, interval, function, args=None, kwargs=None):
        Thread.__init__(self)
        self.interval = interval or LEADER_KEY_LEASE
        self.function = function
        self.args = args or []
        self.kwargs = kwargs or {}
        self.finished = Event()
        self.ready = Event()

    def cancel(self):
        """Stop the timer if it hasn't finished yet"""
        self.finished.set()

    def run(self):
        self.function(*self.args, **self.kwargs)
        while True:
            self.finished.wait(self.interval)
            if not self.finished.is_set():
                self.function(*self.args, **self.kwargs)
                self.ready.set()
            else:
                break


def make_me_leader(identity, server, check_ttl):
    expiry = int(check_ttl * 1000)
    current_leader = server.get(LEADER_KEY)

    if current_leader and current_leader.decode("utf-8") == identity:
        # To avoid expiring key while the leader is still alive, we need
        # to extend the lease if this process is the leader

        logger.debug(f"extending lease to {identity} for {expiry}ms")
        return server.pexpire(LEADER_KEY, expiry)
    else:
        logger.debug(f"setting up leader to {identity} for {expiry}ms")
        return server.set(
            name=LEADER_KEY, value=identity.encode("utf-8"), nx=True, px=expiry
        )


class Election:
    def __init__(self, identity, server, check_ttl=None):
        self.identity = identity
        self.server = server
        self.check_ttl = check_ttl

        self.timer = Timer(
            self.check_ttl, make_me_leader, (identity, server, check_ttl)
        )

    def __enter__(self):
        logger.info(f"starting election process for {self.identity}")
        self.timer.start()

        # The first activation might take some time,
        # so we are waiting a bit before moving on
        self.timer.ready.wait(self.check_ttl * 2)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        logger.info(f"withdrawing from election process for {self.identity}")
        self.timer.cancel()

    def is_leader(self):
        current_leader = self.server.get(LEADER_KEY)
        if current_leader is None:
            logger.warning("no leader, attempting coup")
            make_me_leader(self.identity, self.server, self.check_ttl)
            current_leader = self.server.get(LEADER_KEY)

        if current_leader:
            current_leader = current_leader.decode("utf-8")
        logger.info(f"[{self.identity}] current leader: {current_leader}")
        return current_leader == self.identity



import pytest
from kirby.supervisor.election import Election, make_me_leader
from redis import Redis


@pytest.fixture
def redis():
    redis = Redis()
    redis.flushall()
    yield redis


def test_it_can_make_leader(redis):
    assert make_me_leader("server-1", redis, 0.1)


def test_it_can_elect_a_leader_with_a_single_node(redis):
    with Election("server-1", server=redis, check_ttl=0.1) as bob:
        assert bob.is_leader()


def test_it_can_elect_a_leader_but_not_two(redis):
    with Election("alice", server=redis, check_ttl=0.1) as alice:
        with Election("bob", server=redis, check_ttl=0.1) as bob:
            assert alice.is_leader() ^ bob.is_leader()


def test_it_can_elect_a_new_leader_when_second_stops(redis):
    with Election("alice", server=redis, check_ttl=0.1) as alice:
        assert alice.is_leader()

    redis.flushall()
    with Election("bob", server=redis, check_ttl=0.1) as bob:
        assert bob.is_leader()



import os

import pytest

from kirby.api.queue import Queue
from kirby.exc import CoolDownException
from kirby.supervisor.scheduler import Scheduler
import datetime


def test_scheduler_can_grab_jobs(data_dir):
    scheduler = Scheduler(queue=None, wakeup=60)

    with open(os.path.join(data_dir, "sample_jobs_request.txt"), "r") as f:
        content = f.read()

    jobs = scheduler.parse_jobs(content)

    assert isinstance(jobs, list)
    assert all(isinstance(job, str) for job in jobs)


def test_scheduler_queue_jobs():
    queue = Queue(name="jobs-to-do", testing=True)

    scheduler = Scheduler(queue=queue, wakeup=60)

    scheduler.queue_job("hello world")

    assert queue.last() == "hello world"


def test_scheduler_does_not_queue_twice_within_wakeup_period():
    queue = Queue(name="jobs-to-do", testing=True)

    scheduler = Scheduler(queue=queue, wakeup=30)

    date_1 = datetime.datetime(2000, 1, 1, 0, 0, 0)
    date_2 = datetime.datetime(2000, 1, 1, 0, 0, 20)

    scheduler.queue_job("hello world", now=date_1)
    with pytest.raises(CoolDownException):
        scheduler.queue_job("hello world", now=date_2)


def test_scheduler_can_queue_after_wakeup_period():
    queue = Queue(name="jobs-to-do", testing=True)

    scheduler = Scheduler(queue=queue, wakeup=30)

    date_1 = datetime.datetime(2000, 1, 1, 0, 0, 0)
    date_2 = datetime.datetime(2000, 1, 1, 0, 0, 40)

    scheduler.queue_job("hello world", now=date_1)
    scheduler.queue_job("hello world", now=date_2)



from unittest.mock import patch, MagicMock
import pytest

from kirby.api.ext import WebClient, RETRIES, WAIT_BETWEEN_RETRIES


def test_it_creates_a_kirby_topic(kirby_topic):
    assert not kirby_topic.next()

    kirby_topic.send("Hello world")

    assert kirby_topic.next() == "Hello world"


@pytest.mark.integration
@pytest.mark.skipif(
    not (RETRIES and WAIT_BETWEEN_RETRIES),
    reason="missing EXT_RETRIES and/or WAIT_BETWEEN_RETRIES environment variable",
)
@patch("requests.session")
def test_it_creates_a_kirby_ext_web_client(session_mock):
    data = {"foo": True}

    session_mock.return_value.get.return_value = MagicMock(
        status_code=200, json=MagicMock(return_value=data)
    )

    session_mock.return_value.post.return_value = MagicMock(
        status_code=200, json=MagicMock(return_value={})
    )
    with WebClient(
        "external_server", "http://some.external.server"
    ) as web_client:
        web_client.post("orders", data)
        assert web_client.get("orders") == data


@pytest.mark.integration
@pytest.mark.skipif(
    not (RETRIES and WAIT_BETWEEN_RETRIES),
    reason="missing EXT_RETRIES and/or WAIT_BETWEEN_RETRIES environment variable",
)
@patch("requests.session")
def test_kirby_ext_web_client_handle_get_errors(session_mock):
    data = {"foo": True}

    good_get = MagicMock(status_code=200, json=MagicMock(return_value=data))

    mocking_effects = [
        MagicMock(status_code=i, json=MagicMock(return_value={}))
        for i in [502, 504, 500, 501]
    ]  # Errors
    mocking_effects.append(
        MagicMock(status_code=200, json=MagicMock(return_value={}))
    )  # Empty answer

    session_mock.return_value.post.return_value = MagicMock(
        status_code=200, json=MagicMock(return_value={})
    )
    for mocking_effect in mocking_effects:
        session_mock.return_value.get.side_effect = [mocking_effect, good_get]

        with WebClient(
            "external_server", "http://some.external.server"
        ) as web_client:
            web_client.post("orders", data)
            result = web_client.get("orders")
            assert result == data


@pytest.mark.integration
@pytest.mark.skipif(
    not (RETRIES and WAIT_BETWEEN_RETRIES),
    reason="missing EXT_RETRIES and/or WAIT_BETWEEN_RETRIES environment variable",
)
@patch("requests.session")
def test_kirby_ext_web_client_handle_post_errors(session_mock):
    data = {"foo": True}
    session_mock.return_value.get.return_value = MagicMock(
        status_code=200, json=MagicMock(return_value=data)
    )
    session_mock.return_value.post.side_effect = [
        MagicMock(status_code=500, json=MagicMock(return_value={})),
        MagicMock(status_code=200, json=MagicMock(return_value={})),
    ]
    with WebClient(
        "external_server", "http://some.external.server"
    ) as web_client:
        web_client.post("orders", data)
        result = web_client.get("orders")
        assert result == data



from flask_security import SQLAlchemyUserDatastore, Security
from flask_security import RoleMixin, UserMixin

from . import db

from enum import Enum


class UserRoles(Enum):
    ADMIN = "admin"


class ProviderTypes(Enum):
    LOCAL = "local"
    LDAP = "ldap"


roles_users = db.Table(
    "roles_users",
    db.Column("user_id", db.Integer(), db.ForeignKey("user.id")),
    db.Column("role_id", db.Integer(), db.ForeignKey("role.id")),
)


class Role(db.Model, RoleMixin):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(100), unique=True)
    description = db.Column(db.String(255))

    def __repr__(self):
        return f"<Role: {self.name}>"


class User(db.Model, UserMixin):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(255), unique=True)
    email = db.Column(db.String(255))
    password = db.Column(db.String(255))
    active = db.Column(db.Boolean(), default=True)
    provider = db.Column(db.Enum(ProviderTypes), default=ProviderTypes.LOCAL)
    roles = db.relationship(
        "Role",
        secondary=roles_users,
        backref=db.backref("users", lazy="dynamic"),
    )

    @property
    def is_local(self):
        return self.provider == ProviderTypes.LOCAL

    def has_role(self, role):
        if isinstance(role, UserRoles):
            return self.has_role(role=role.value)
        if isinstance(role, str):
            role = role.lower()
            return role in (role.name.lower() for role in self.roles)
        else:
            return role in self.roles

    @property
    def is_admin(self):
        return self.has_role(UserRoles.ADMIN)


user_datastore = SQLAlchemyUserDatastore(db, User, Role)
security = Security()



import datetime
import json

import requests
from smart_getenv import getenv

import logging

from ..exc import CoolDownException

logger = logging.getLogger(__name__)


class Scheduler:
    def __init__(self, queue, wakeup):
        self.queue = queue
        self.cooldown = datetime.timedelta(seconds=wakeup)

    def fetch_jobs(self):
        url = getenv("KIRBY_SCHEDULE_ENDPOINT", type=str)
        try:
            response = requests.get(url)
            return response.text
        except requests.exceptions.ConnectionError:
            logger.exception("unable to fetch jobs: ")

    def parse_jobs(self, content):
        jobs = []
        descriptions = json.loads(content)
        for description in descriptions["jobs"]:
            jobs.append(json.dumps(description))

        return jobs

    def queue_job(self, job, now=None):
        if now is None:
            now = datetime.datetime.utcnow()

        submitted_jobs = self.queue.between(start=now - self.cooldown, end=now)

        if job in submitted_jobs:
            raise CoolDownException()
        else:
            self.queue.append(job, submitted=now)



import os
from smart_getenv import getenv
from pytest import fixture
from freezegun import freeze_time

from kirby.api import Kirby
from kirby.api.ext import Topic
from tests.conftest import API_ROOT


DATE = "2019-05-22 15:18"
TOPIC_NAME = "orders"


@fixture
def kirby_hidden_env(db_scripts_not_registered):
    env = {
        # These are the "hidden" variables.
        "ID": "1",
        "PACKAGE_NAME": "orders_retriever",
        "KIRBY_WEB_SERVER": API_ROOT,
        # "KAFKA_BOOTSTRAP_SERVERS": KAFKA_BOOTSTRAP_SERVERS,
        # ---
        # These are the one defined into the signature when Kirby is called.
        # They are specifics to the script.
        "WEBCLIENT_ENDPOINT": "http://a.place.somewhere",
        "TOPIC_NAME": TOPIC_NAME,
    }

    for k, v in env.items():
        os.environ[k] = v
    return env


@fixture
def kirby_expected_env():
    return {"WEBCLIENT_ENDPOINT": {"type": str}, "TOPIC_NAME": {"type": str}}


@fixture(scope="function")
@freeze_time(DATE)
def kirby_app(session, kirby_hidden_env, kirby_expected_env):
    return Kirby(kirby_expected_env, session=session)


@fixture(scope="function")
def kirby_topic(kirby_app, kafka_topic_factory):
    import logging

    logger = logging.getLogger(__name__)
    bootstrap_servers = getenv(
        "KAFKA_BOOTSTRAP_SERVERS", type=list, separator=","
    )
    if bootstrap_servers:
        with kafka_topic_factory(TOPIC_NAME):
            with Topic(kirby_app, "TOPIC_NAME") as topic:
                yield topic
    else:
        logger.warning(
            f"There is no KAFKA_BOOTSTRAP_SERVERS. "
            f"Topic will be created in testing mode"
        )
        with Topic(kirby_app, "TOPIC_NAME", testing=True) as topic:
            yield topic



import attr
import json


def convert_variables(data):
    export = {}
    for variable in data:
        export[variable["key"]] = variable["value"]
    return export


@attr.s(hash=True)
class JobDescription:
    name = attr.ib(type=str)
    environment = attr.ib(type=str)
    package_name = attr.ib(type=str)
    package_version = attr.ib(type=str)
    notifications = attr.ib(type=tuple)
    variables = attr.ib(type=dict, convert=convert_variables)


def parse_job_description(job_description):
    kwargs = json.loads(job_description)
    job = JobDescription(**kwargs)

    return job



from datetime import datetime
from sqlalchemy.orm.exc import NoResultFound
from flask import abort
from cronex import CronExpression
import copy
from flask_restplus import Api, Resource, reqparse, fields

from ..models import db, Script, Topic, Job, JobType

api = Api()

registration_parser = reqparse.RequestParser()
registration_parser.add_argument("script_id", type=int)
registration_parser.add_argument(
    "source_id", type=int, action="append", default=[]
)
registration_parser.add_argument(
    "destination_id", type=int, action="append", default=[]
)


@api.route("/registration")
class Registration(Resource):
    @api.expect(registration_parser)
    def patch(self):
        args = registration_parser.parse_args()

        # Update last_seen
        try:
            script = (
                db.session.query(Script).filter_by(id=args["script_id"]).one()
            )
            script.last_seen = datetime.utcnow()
        except NoResultFound:
            abort(
                400,
                f"The script with the 'id':{args['script_id']}, doesn't exist.",
            )

        # Add sources and destinations
        for source_id in args["source_id"]:
            try:
                source = db.session.query(Topic).filter_by(id=source_id).one()
                script.add_source(source)
            except NoResultFound:
                db.session.rollback()
                abort(
                    400,
                    f"The topic with the 'id':{source_id}, doesn't exist. "
                    f"No changes has been committed.",
                )
        for destination_id in args["destination_id"]:
            try:
                destination = (
                    db.session.query(Topic).filter_by(id=destination_id).one()
                )
                script.add_destination(destination)
            except NoResultFound:
                db.session.rollback()
                abort(
                    400,
                    f"The topic with the 'id':{destination_id}, doesn't exist. "
                    f"No changes has been committed.",
                )

        db.session.commit()

        return {"message": "Sucess"}


topic_parser = reqparse.RequestParser()
topic_parser.add_argument("name", type=str, required=True)


@api.route("/topic")
class TopicView(Resource):
    @api.expect(topic_parser)
    def get(self):
        args = topic_parser.parse_args()
        topic = db.session.query(Topic).filter_by(name=args["name"]).one()
        return {"id": topic.id, "name": topic.name}


group_model = api.model(
    "Group", {"name": fields.String, "emails": fields.List(fields.String)}
)

notifications_model = api.model(
    "Notifications",
    {
        "on_retry": fields.Boolean,
        "on_failure": fields.Boolean,
        "groups": fields.List(fields.Nested(group_model)),
    },
)

config_model = api.model(
    "Config", {"key": fields.String, "value": fields.String}
)


job_model = api.model(
    "Job",
    {
        "name": fields.String,
        "environment": fields.String,
        "package_name": fields.String,
        "package_version": fields.String,
        "notifications": fields.List(fields.Nested(notifications_model)),
        "variables": fields.List(fields.Nested(config_model)),
    },
)

schedule_model = api.model(
    "Schedule",
    {
        "date": fields.String(default=str(datetime.utcnow())),
        "jobs": fields.List(fields.Nested(job_model)),
    },
)


def should_run(schedule, date):
    cron = CronExpression(
        " ".join(
            [
                schedule.minute or "*",
                schedule.hour or "*",
                schedule.day or "*",
                schedule.month or "*",
                schedule.weekday or "*",
            ]
        )
    )
    now = [date.year, date.month, date.day, date.hour, date.minute]
    return cron.check_trigger(now)


@api.route("/schedule")
class Schedule(Resource):
    @api.marshal_with(schedule_model)
    def get(self):
        date = datetime.utcnow()
        jobs = []
        schedule = {"date": date, "jobs": jobs}

        for job in db.session.query(Job).all():
            job_name = job.name

            # Get Notifications data
            notifications = []
            for notification in job.notifications:
                on_retry = notification.on_retry
                on_failure = notification.on_failure
                groups = [
                    {
                        "name": group.name,
                        "emails": [
                            group_email.email for group_email in group.emails
                        ],
                    }
                    for group in notification.groups
                ]
                notifications.append(
                    {
                        "on_retry": on_retry,
                        "on_failure": on_failure,
                        "groups": copy.deepcopy(groups),
                    }
                )

            # Each Job can have various contexts. We need to go through them
            # to generate all the jobs. And for each context, linked scripts
            # contains the package_name and package_version are in the Context
            # data.
            for context in job.contexts:
                if job.type == JobType.TRIGGERED or (
                    context.schedules
                    and any(
                        [
                            should_run(schedule, date)
                            for schedule in context.schedules
                        ]
                    )
                ):
                    for script in context.scripts:
                        jobs.append(
                            {
                                "name": job_name,
                                "environment": context.environment.name,
                                "package_name": script.package_name,
                                "package_version": script.package_version,
                                "notifications": copy.deepcopy(notifications),
                                "variables": [
                                    {"key": k, "value": v}
                                    for k, v in context.variables().items()
                                ],
                            }
                        )

        return schedule



import os
from freezegun import freeze_time
from datetime import datetime
import pytest
from unittest.mock import patch, MagicMock

from tests.tests_api.conftest import DATE

from kirby.models import db, Script, Topic
from kirby.api import Kirby, ClientError, ServerError


def get_script_in_db_from_id(id_script):
    return db.session.query(Script).filter_by(id=id_script).one()


def get_topic_in_db_from_name(name_topic):
    return db.session.query(Topic).filter_by(name=name_topic).one()


@freeze_time(DATE)
def test_it_create_a_kirby_app(kirby_app, kirby_hidden_env):
    assert (
        kirby_app.ctx.WEBCLIENT_ENDPOINT
        == kirby_hidden_env["WEBCLIENT_ENDPOINT"]
    )

    script_in_db = get_script_in_db_from_id(kirby_app.ctx.ID)
    assert script_in_db.last_seen == datetime.utcnow()


@pytest.mark.integration
@pytest.mark.skipif(
    not os.getenv("KAFKA_BOOTSTRAP_SERVERS"),
    reason="missing KAFKA_BOOTSTRAP_SERVERS environment",
)
def test_it_add_source(
    kirby_app, kirby_topic, db_scripts_not_registered, db_topics
):
    kirby_app.add_source(kirby_topic)

    script_in_db = get_script_in_db_from_id(kirby_app.ctx.ID)

    assert script_in_db.sources[0] == get_topic_in_db_from_name(
        kirby_topic.name
    )


@pytest.mark.integration
@pytest.mark.skipif(
    not os.getenv("KAFKA_BOOTSTRAP_SERVERS"),
    reason="missing KAFKA_BOOTSTRAP_SERVERS environment",
)
def test_it_add_destination(
    session, kirby_app, kirby_topic, db_scripts_not_registered, db_topics
):
    kirby_app.add_destination(kirby_topic)

    script_in_db = get_script_in_db_from_id(kirby_app.ctx.ID)

    assert script_in_db.destinations[0] == get_topic_in_db_from_name(
        kirby_topic.name
    )


def test_throw_error_if_bad_usage(kirby_app):
    with pytest.raises(ClientError):
        kirby_app._register(source_id=-1, destination_id=-1)


@patch("requests.session")
def test_throw_error_at_init_if_server_error(
    session_mock, kirby_expected_env, kirby_hidden_env
):
    session_mock.return_value.patch.return_value = MagicMock(
        status_code=500, json=MagicMock(return_value={})
    )

    # Kirby register itself at init, at least to update 'last_seen' in the DB
    with pytest.raises(ServerError):
        Kirby(kirby_expected_env)


@patch("requests.session")
def test_throw_error_at_source_registration_if_error(
    session_mock, kirby_expected_env, kirby_hidden_env
):
    session_mock.return_value.get.side_effect = [
        MagicMock(
            status_code=500, json=MagicMock(return_value={"Server Error"})
        ),
        MagicMock(
            status_code=400, json=MagicMock(return_value={"Client Error"})
        ),
    ]
    session_mock.return_value.patch.return_value = MagicMock(
        status_code=200, json=MagicMock(return_value={})
    )

    kirby_app = Kirby(kirby_expected_env)
    with pytest.raises(ServerError):
        kirby_app.add_destination(MagicMock(id=1))
    with pytest.raises(ClientError):
        kirby_app.add_destination(MagicMock(id=-1))



import msgpack
from kafka import KafkaProducer, KafkaConsumer
from smart_getenv import getenv
import logging
import tenacity

from .ext import Topic, kafka_retry_args

logger = logging.getLogger(__name__)


class Queue(Topic):
    def __init__(self, name, testing=False, ssl_security_protocol=True):
        self.name = name
        self.testing = testing
        self.ssl_security_protocol = ssl_security_protocol
        self.init_kafka()
        mode = "testing" if self.testing else "live"
        logger.debug(f"starting queue {self.name} in {mode} mode")

    @tenacity.retry(**kafka_retry_args)
    def init_kafka(self):
        if self.testing:
            self._messages = []
        else:
            bootstrap_servers = getenv(
                "KAFKA_BOOTSTRAP_SERVERS", type=list, separator=","
            )
            kafka_args = {"bootstrap_servers": bootstrap_servers}

            if self.ssl_security_protocol:
                kafka_args.update(
                    {
                        "ssl_cafile": getenv("KAFKA_SSL_CAFILE", type=str),
                        "ssl_certfile": getenv("KAFKA_SSL_CERTFILE", type=str),
                        "ssl_keyfile": getenv("KAFKA_SSL_KEYFILE", type=str),
                        "security_protocol": "SSL",
                    }
                )

            logger.debug(f"bootstrap servers: {bootstrap_servers}")

            self._producer = KafkaProducer(
                value_serializer=msgpack.dumps, **kafka_args
            )

            self._consumer = KafkaConsumer(
                self.name,
                group_id=getenv("KIRBY_SUPERVISOR_GROUP_ID", type=str),
                enable_auto_commit=True,
                value_deserializer=msgpack.loads,
                **kafka_args,
            )

    def append(self, *args, **kargs):
        super().send(*args, **kargs)

    def last(self):
        if self.testing:
            _, msg = self._messages[-1]
            return msg
        else:
            raise NotImplementedError("this is only for testing")



import logging
import datetime
import requests
import msgpack
from smart_getenv import getenv
import tenacity

from kafka import KafkaConsumer, KafkaProducer
from kafka.errors import NoBrokersAvailable

logger = logging.getLogger(__name__)

RETRIES = getenv("EXT_RETRIES", type=int)
WAIT_BETWEEN_RETRIES = getenv("EXT_WAIT_BETWEEN_RETRIES", type=float)
kafka_retry_args = {
    "retry": tenacity.retry_if_exception_type(NoBrokersAvailable),
    "wait": tenacity.wait_fixed(WAIT_BETWEEN_RETRIES),
    "stop": tenacity.stop_after_attempt(RETRIES),
    "reraise": True,
}


class WebClientError(Exception):
    pass


webserver_retry_args = {
    "retry": tenacity.retry_if_exception_type(WebClientError),
    "wait": tenacity.wait_fixed(WAIT_BETWEEN_RETRIES),
    "stop": tenacity.stop_after_attempt(RETRIES),
    "reraise": True,
}


class Topic:
    def __init__(
        self,
        kirby_app,
        topic_name_variable_name,
        ssl_security_protocol=True,
        testing=False,
    ):
        self.name = kirby_app.ctx[topic_name_variable_name]
        self.testing = testing
        self.ssl_security_protocol = ssl_security_protocol
        self.kirby_app = kirby_app
        self.init_kafka()
        mode = "testing" if self.testing else "live"
        logger.debug(f"starting topic {self.name} in {mode} mode")

    @tenacity.retry(**kafka_retry_args)
    def init_kafka(self):
        kafka_args = {
            "bootstrap_servers": self.kirby_app.ctx.KAFKA_BOOTSTRAP_SERVERS
        }

        if self.ssl_security_protocol:
            kafka_args.update(
                {
                    "ssl_cafile": self.kirby_app.ctx.KAFKA_SSL_CAFILE,
                    "ssl_certfile": self.kirby_app.ctx.KAFKA_SSL_CERTFILE,
                    "ssl_keyfile": self.kirby_app.ctx.KAFKA_SSL_KEYFILE,
                    "security_protocol": "SSL",
                }
            )
        if self.testing:
            self._messages = []
            self.cursor_position = 0
        else:
            self._consumer = KafkaConsumer(
                self.name,
                group_id=self.kirby_app.ctx.PACKAGE_NAME,
                value_deserializer=lambda x: msgpack.loads(x, raw=False),
                **kafka_args,
            )
            self._producer = KafkaProducer(
                value_serializer=msgpack.dumps, **kafka_args
            )

    @tenacity.retry(**kafka_retry_args)
    def send(self, message, submitted=None):
        if submitted is None:
            submitted = datetime.datetime.utcnow()

        if self.testing:
            self._messages.append((submitted, message))

        else:
            timestamp_ms = int(submitted.timestamp() * 1000)
            self._producer.send(
                self.name, value=message, timestamp_ms=timestamp_ms
            )
            self._producer.flush()

    @tenacity.retry(**kafka_retry_args)
    def between(self, start, end):
        if self.testing:
            return [msg for t, msg in self._messages if start <= t < end]
        else:
            self._consumer.poll(timeout_ms=5000)

            partitions = self._consumer.assignment()

            start_mapping = {p: start.timestamp() for p in partitions}
            start_offsets = self._consumer.offsets_for_times(start_mapping)

            start_timestamp = start.timestamp() * 1000
            end_timestamp = end.timestamp() * 1000

            # with modify_temporarily_offsets(self._consumer, partitions):
            messages = []
            for partition, offsets in start_offsets.items():
                self._consumer.seek(partition=partition, offset=offsets.offset)

            records_by_partition = self._consumer.poll(timeout_ms=5000)

            for partition, records in records_by_partition.items():
                for record in records:
                    if start_timestamp <= record.timestamp < end_timestamp:
                        messages.append((record.timestamp, record.value))

            messages.sort(key=lambda x: x[0])
            return [v for t, v in messages]

    @staticmethod
    def message_to_item(raw_message):
        if raw_message:
            for messages_by_topic in raw_message.values():
                for message in messages_by_topic:
                    return message.value

    @tenacity.retry(**kafka_retry_args)
    def next(self, timeout_ms=500):
        if not self.testing:
            message = self._consumer.poll(max_records=1, timeout_ms=timeout_ms)
            self._consumer.commit()
            return self.message_to_item(message)
        else:
            if self._messages:
                (_, message) = self._messages[self.cursor_position]
                self.cursor_position += 1
                return message
            else:
                return None

    def close(self):
        if not self.testing:
            self._producer.close()
            self._consumer.close(autocommit=False)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class WebClient:
    def __init__(self, name, web_endpoint_base, session=None):
        self.name = name
        self.web_endpoint_base = web_endpoint_base
        self._session = session or requests.session()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @tenacity.retry(**webserver_retry_args)
    def post(self, endpoint, data, params=None):
        if not params:
            params = {}
        result = self._session.post(
            "/".join([self.web_endpoint_base, endpoint]),
            data=data,
            params=params,
        )
        if result.status_code == 200:
            return result.json()
        raise WebClientError(
            f"POST error on {result.url}. "
            f"Status code : {result.status_code}. "
            f"Response : {result.text}"
        )

    @tenacity.retry(**webserver_retry_args)
    def get(self, endpoint, params=None):
        result = self._session.get(
            "/".join([self.web_endpoint_base, endpoint]), params=params
        )
        if result.status_code == 200:
            json = result.json()
            if json:
                return json
        raise WebClientError(
            f"GET error on {result.url}. "
            f"Status code : {result.status_code}. "
            f"Response : {result.text}"
        )



from os import name

from kirby.models import ConfigKey
from kirby.models.security import user_datastore
from kirby import models


def create_demo_db(s):

    # create demo user
    user = user_datastore.create_user(username="demo", password="demo")
    role = user_datastore.find_role("admin")
    user_datastore.add_role_to_user(user=user, role=role)
    s.commit()

    sysadmins = models.NotificationGroup(name="sysadmins")
    sysadmins.add_email("alice@local.local")
    sysadmins.add_email("bob@local.local")
    sysadmins.add_email("charles@local.local")

    s.add(sysadmins)
    s.commit()

    # fill the database with dummy content
    job = models.Job(
        name="Fetch bakery realtime sales", type=models.JobType.SCHEDULED
    )
    s.add(job)
    job.add_notification(sysadmins, on_failure=True, on_retry=False)

    job.set_config(
        SENTRY_DSN="http://sentry.dsn.somewhere", SSH_USERNAME="demo"
    )

    dev = models.Environment(name="Development")
    test = models.Environment(name="Test")
    prod = models.Environment(name="Production")

    s.add_all([dev, test, prod])

    dev_ctx = models.Context(job=job, environment=dev)
    dev_ctx.set_config(ENV="dev", SSH_SERVER="dev.server.somewhere:22")
    test_ctx = models.Context(job=job, environment=test)
    test_ctx.set_config(ENV="test", SSH_SERVER="test.server.somewhere:25")
    prod_ctx = models.Context(job=job, environment=prod)
    prod_ctx.set_config(ENV="prod", SSH_SERVER="server.somewhere:22")
    s.add_all([dev_ctx, test_ctx, prod_ctx])

    schedule = models.Schedule(
        name="every two minutes", hour="*", minute="*/2"
    )
    s.add(schedule)
    dev_ctx.add_schedule(schedule)
    test_ctx.add_schedule(schedule)
    prod_ctx.add_schedule(schedule)

    source = models.Topic(name="bakery")
    s.add(source)
    destination = models.Topic(name="timeseries")
    s.add(destination)

    dev_script = models.Script(
        context=dev_ctx,
        package_name="test_package_for_dev",
        package_version="2.0.1",
    )
    dev_script.add_source(source)
    dev_script.add_destination(destination)

    test_script = models.Script(
        context=dev_ctx,
        package_name="test_package_for_test",
        package_version="1.0.1",
    )
    test_script.add_source(source)
    test_script.add_destination(destination)

    prod_script = models.Script(
        context=dev_ctx,
        package_name="test_package_for_prod",
        package_version="0.0.5",
    )
    prod_script.add_source(source)
    prod_script.add_destination(destination)

    s.add_all([dev_script, test_script, prod_script])

    s.add(ConfigKey(name="KAFKA_URL", value="some.kafka.server:9999"))

    s.commit()



from enum import Enum
from cronex import CronExpression

from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.orm import Session, validates
from sqlalchemy import UniqueConstraint, event

db = SQLAlchemy()


class Environment(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(), nullable=False, unique=True)

    def __repr__(self):
        return self.name


class JobType(Enum):
    SCHEDULED = "scheduled"
    TRIGGERED = "triggered"


class Job(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(), nullable=False, unique=True)
    description = db.Column(db.Text())
    type = db.Column(db.Enum(JobType), nullable=False)

    def add_context(self, context):
        self.contexts.append(context)

    def add_notification(self, group, on_retry=False, on_failure=True):
        notification = Notification(on_retry=on_retry, on_failure=on_failure)
        notification.groups.append(group)
        self.notifications.append(notification)

    def set_config(self, **kwargs):
        for key, value in kwargs.items():
            c = ConfigKey(name=key, value=str(value))
            self.config_keys.append(c)

    def __repr__(self):
        return f"{self.name} ({self.type.value})"


schedules_to_contexts = db.Table(
    "schedules_to_contexts",
    db.metadata,
    db.Column("schedule", db.Integer, db.ForeignKey("schedule.id")),
    db.Column("context", db.Integer, db.ForeignKey("context.id")),
)


class Context(db.Model):
    id = db.Column(db.Integer, primary_key=True)

    environment_id = db.Column(
        db.Integer, db.ForeignKey("environment.id"), nullable=False
    )
    environment = db.relationship(
        Environment, backref=db.backref("contexts", lazy=True)
    )

    job_id = db.Column(db.Integer, db.ForeignKey("job.id"), nullable=False)
    job = db.relationship(Job, backref=db.backref("contexts", lazy=True))

    schedules = db.relationship(
        "Schedule", secondary=schedules_to_contexts, back_populates="contexts"
    )

    def set_config(self, **kwargs):
        for key, value in kwargs.items():
            c = ConfigKey(name=key, value=str(value))
            self.config_keys.append(c)

    def add_schedule(self, schedule):
        self.schedules.append(schedule)

    def __repr__(self):
        return f"{self.job} @ {self.environment}"

    def variables(self):
        export = {}

        global_query = db.session.query(ConfigKey).filter_by(
            scope=ConfigScope.GLOBAL
        )
        job_query = db.session.query(ConfigKey).filter(
            ConfigKey.job_id == self.job.id
        )
        own_query = db.session.query(ConfigKey).filter(
            ConfigKey.context_id == self.id
        )

        query = global_query.union(job_query).union(own_query)

        for config in query.all():
            export[config.name] = config.value

        return export


class ConfigScope(Enum):
    GLOBAL = "global"
    JOB = "job"
    CONTEXT = "context"


class ConfigKey(db.Model):
    __table_args__ = (
        UniqueConstraint(
            "name", "context_id", "job_id", name="_context_name_uc"
        ),
    )

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(), nullable=False)
    value = db.Column(db.String(), nullable=False)
    scope = db.Column(db.Enum(ConfigScope), nullable=False)

    context_id = db.Column(
        db.Integer, db.ForeignKey("context.id"), nullable=True
    )
    context = db.relationship(
        Context, backref=db.backref("config_keys", lazy=True)
    )

    job_id = db.Column(db.Integer, db.ForeignKey("job.id"), nullable=True)
    job = db.relationship(Job, backref=db.backref("config_keys", lazy=True))

    @property
    def job_name(self):  # pragma: no cover
        if self.job:
            return self.job.name
        elif self.context:
            return self.context.job.name

    @property
    def environment_name(self):  # pragma: no cover
        if self.context:
            return self.context.environment.name


@event.listens_for(Session, "before_flush")
def set_config_scope(session, flush_context, instances):
    for target in set(session.new).union(set(session.dirty)):
        if isinstance(target, ConfigKey):
            if target.context or target.context_id:
                target.scope = ConfigScope.CONTEXT
            elif target.job or target.job_id:
                target.scope = ConfigScope.JOB
            else:
                target.scope = ConfigScope.GLOBAL


class Schedule(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(), nullable=False, unique=True)
    minute = db.Column(db.String(), default="*")
    hour = db.Column(db.String(), default="*")
    day = db.Column(db.String(), default="*")
    month = db.Column(db.String(), default="*")
    weekday = db.Column(db.String(), default="*")

    contexts = db.relationship(
        "Context", secondary=schedules_to_contexts, back_populates="schedules"
    )

    def add_suspension(self, suspension):
        self.suspensions.append(suspension)

    def __repr__(self):
        return self.name

    @validates("hour", "minute", "day", "month", "weekday")
    def validate_schedule_attribute(self, key, attribute):
        cron_expression = "{minute} {hour} {day} {month} {weekday}".format(
            minute=attribute if key == "minute" else "*",
            hour=attribute if key == "hour" else "*",
            day=attribute if key == "day" else "*",
            month=attribute if key == "month" else "*",
            weekday=attribute if key == "weekday" else "*",
        )

        try:
            CronExpression(cron_expression)
        except ValueError:
            raise ValueError(
                f"The Schedule cannot accept the "
                f"value given in the {key} attribute"
            )

        return attribute


class Suspension(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    start = db.Column(db.DateTime, nullable=False)
    end = db.Column(db.DateTime, nullable=False)
    reason = db.Column(db.Text(), nullable=True)

    schedule_id = db.Column(
        db.Integer, db.ForeignKey("schedule.id"), nullable=False
    )
    schedule = db.relationship(
        Schedule, backref=db.backref("suspensions", lazy=True)
    )


notification_to_groups = db.Table(
    "notification_to_groups",
    db.metadata,
    db.Column("notification", db.Integer, db.ForeignKey("notification.id")),
    db.Column(
        "notification_group",
        db.Integer,
        db.ForeignKey("notification_group.id"),
    ),
)


class NotificationGroup(db.Model):
    __tablename__ = "notification_group"
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(), nullable=False, unique=True)

    notifications = db.relationship(
        "Notification",
        secondary=notification_to_groups,
        back_populates="groups",
    )

    def add_email(self, email):
        ne = NotificationEmail(email=email)
        self.emails.append(ne)

    def __repr__(self):
        return self.name


class NotificationEmail(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    email = db.Column(db.String(), nullable=False)

    notification_group_id = db.Column(
        db.Integer, db.ForeignKey("notification_group.id"), nullable=False
    )
    notification_group = db.relationship(
        NotificationGroup, backref=db.backref("emails", lazy=True)
    )


class Notification(db.Model):
    id = db.Column(db.Integer, primary_key=True)

    on_retry = db.Column(db.Boolean, nullable=False, default=False)
    on_failure = db.Column(db.Boolean, nullable=False, default=True)

    job_id = db.Column(db.Integer, db.ForeignKey("job.id"), nullable=False)
    job = db.relationship(Job, backref=db.backref("notifications", lazy=True))

    groups = db.relationship(
        "NotificationGroup",
        secondary=notification_to_groups,
        back_populates="notifications",
    )

    def __repr__(self):
        return f"{self.job} -> {','.join(g.name for g in self.groups)}"


class Script(db.Model):
    __tablename__ = "script"
    id = db.Column(db.Integer, primary_key=True)
    package_name = db.Column(db.String(), nullable=False)
    package_version = db.Column(db.String(), nullable=False)

    context_id = db.Column(
        db.Integer, db.ForeignKey("context.id"), nullable=False
    )
    context = db.relationship(
        Context, backref=db.backref("scripts", lazy=True)
    )

    first_seen = db.Column(db.DateTime)
    last_seen = db.Column(db.DateTime)

    def add_source(self, source):
        self.sources.append(source)

    def add_destination(self, destination):
        self.destinations.append(destination)

    def __repr__(self):
        return f"{self.package_name} / {self.package_version}"


class Topic(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(), nullable=False, unique=True)

    subscriber_id = db.Column(db.Integer, db.ForeignKey("script.id"))
    subscriber = db.relationship(
        Script,
        backref=db.backref("sources", lazy=True),
        foreign_keys=[subscriber_id],
    )

    provider_id = db.Column(db.Integer, db.ForeignKey("script.id"))
    provider = db.relationship(
        Script,
        backref=db.backref("destinations", lazy=True),
        foreign_keys=[provider_id],
    )

    def __repr__(self):
        return self.name



import datetime
import multiprocessing
import os

import pytest

from kirby.api.context import ContextManager
from kirby.api.queue import Queue


def _load_config(q):
    from kirby.api.context import ctx

    assert ctx.HELLO == "WORLD"
    assert ctx.MYLIST == ["this", "is", "a", "list"]
    assert ctx["HELLO"]

    q.put(True)


def test_it_can_read_configuration():
    os.environ["HELLO"] = "WORLD"
    os.environ["MYLIST"] = "this:is:a:list"

    ContextManager(
        {"HELLO": {"type": str}, "MYLIST": {"type": list, "separator": ":"}}
    )

    from kirby.api.context import ctx

    assert ctx.HELLO == "WORLD"
    assert ctx.MYLIST == ["this", "is", "a", "list"]

    q = multiprocessing.Queue(maxsize=1)
    ps = multiprocessing.Process(target=_load_config, args=(q,))
    ps.start()

    assert q.get(block=True, timeout=2)


def test_it_can_create_a_queue():
    q = Queue("my-queue", testing=True)
    q.append("hello")
    assert q.last() == "hello"


@pytest.mark.integration
@pytest.mark.skipif(
    not os.getenv("KAFKA_BOOTSTRAP_SERVERS"),
    reason="missing KAFKA_BOOTSTRAP_SERVERS environment",
)
def test_it_can_create_a_queue_integration(kafka_topic_factory):
    offset = datetime.timedelta(seconds=5)

    with kafka_topic_factory("kirby-test-integration"):
        q = Queue("kirby-test-integration")

        start = datetime.datetime.now()

        q.append("too early", submitted=start - offset)
        q.append("hello world", submitted=start + offset)
        q.append("too late", submitted=start + offset + offset)

        messages = q.between(start, start + offset)

        assert messages == [b"hello world"]



import logging
from time import perf_counter
from time import sleep

from redis import Redis
from smart_getenv import getenv

from .election import Election
from .scheduler import Scheduler
from ..api.queue import Queue

logger = logging.getLogger(__name__)


def run_supervisor(name, window, wakeup):
    server = Redis()
    queue = Queue(
        name=getenv(
            "KIRBY_TOPIC_JOB_OFFERS", type=str, default=".kirby.job-offers"
        )
    )
    scheduler = Scheduler(queue=queue, wakeup=wakeup)
    with Election(identity=name, server=server, check_ttl=window) as me:
        while True:
            checkpoint = perf_counter()
            if me.is_leader():
                content = scheduler.fetch_jobs()
                if content is not None:
                    jobs = scheduler.parse_jobs(content)
                    for job in jobs:
                        scheduler.queue_job(job)
            else:
                logger.debug("not the leader, do nothing")

            drift = perf_counter() - checkpoint
            next_wakeup = wakeup - drift
            logger.debug("waking up in {:.2f}s".format(next_wakeup))
            sleep(next_wakeup)



from flask import Flask, url_for
from flask_admin import helpers as admin_helpers
from smart_getenv import getenv

from .admin import admin
from .endpoints import api
from .forms import LoginForm
from ..models import db
from ..models.security import user_datastore, security, UserRoles


def app_maker(config=None):
    app = Flask("kirby")

    app.config["TESTING"] = getenv("TESTING", type=bool)
    app.config["SQLALCHEMY_DATABASE_URI"] = getenv("SQLALCHEMY_DATABASE_URI")
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = getenv(
        "SQLALCHEMY_TRACK_MODIFICATIONS"
    )
    app.config["SECRET_KEY"] = getenv("SECRET_KEY")
    app.config["SECURITY_PASSWORD_SALT"] = getenv("SECURITY_PASSWORD_SALT")

    if config:  # pragma: no cover
        app.config.update(config)

    db.init_app(app)
    admin.init_app(app)
    api.init_app(app)
    security._state = security.init_app(
        app=app, datastore=user_datastore, login_form=LoginForm
    )

    @security.context_processor
    def security_context_processor():
        return dict(
            admin_base_template=admin.base_template,
            admin_view=admin.index_view,
            h=admin_helpers,
            get_url=url_for,
        )

    @app.before_first_request
    def create_models():
        db.create_all()
        user_datastore.find_or_create_role(UserRoles.ADMIN.value)
        db.session.commit()

    return app



from flask import current_app, request
from flask_security.confirmable import requires_confirmation
from flask_security.forms import (
    NextFormMixin,
    get_form_field_label,
    config_value,
)
from flask_security.utils import get_message, verify_and_update_password
from flask_wtf import FlaskForm
from werkzeug.local import LocalProxy
from wtforms import BooleanField, PasswordField, StringField, SubmitField


_security = LocalProxy(lambda: current_app.extensions["security"])
_datastore = LocalProxy(lambda: current_app.extensions["security"].datastore)

import logging

logger = logging.getLogger(__name__)


class LoginForm(FlaskForm, NextFormMixin):
    """Username login form"""

    username = StringField(get_form_field_label("username"))
    password = PasswordField(get_form_field_label("password"))
    remember = BooleanField(get_form_field_label("remember_me"))
    submit = SubmitField(get_form_field_label("login"))

    def __init__(self, *args, **kwargs):
        super(LoginForm, self).__init__(*args, **kwargs)
        if not self.next.data:
            self.next.data = request.args.get("next", "")
        self.remember.default = config_value("DEFAULT_REMEMBER_ME")

    def validate(self):
        if not super(LoginForm, self).validate():
            return False

        if not self.username.data.strip():
            self.username.errors.append("Username not provided")
            return False

        if not self.password.data.strip():
            self.password.errors.append(
                get_message("PASSWORD_NOT_PROVIDED")[0]
            )
            return False

        username = self.username.data
        self.user = _security.datastore.find_user(username=username)
        if not self.user:
            logger.warning(
                "not found {} using username field, "
                "now using fallback with email".format(username)
            )
            self.user = _security.datastore.find_user(email=username)

        if self.user is None:
            self.username.errors.append(get_message("USER_DOES_NOT_EXIST")[0])
            return False
        if not self.user.password:
            self.password.errors.append(get_message("PASSWORD_NOT_SET")[0])
            return False
        if not verify_and_update_password(self.password.data, self.user):
            self.password.errors.append(get_message("INVALID_PASSWORD")[0])
            return False
        if requires_confirmation(self.user):
            self.username.errors.append(
                get_message("CONFIRMATION_REQUIRED")[0]
            )
            return False
        if not self.user.is_active:
            self.username.errors.append(get_message("DISABLED_ACCOUNT")[0])
            return False
        return True



import os
from kirby.supervisor.executor import parse_job_description


def test_executor_can_parse_job(data_dir):

    with open(os.path.join(data_dir, "sample_single_job.txt"), "r") as f:
        job_description = f.read()

    job = parse_job_description(job_description)

    assert job.name == "Fetch bakery realtime sales"
    assert job.environment == "Development"
    assert job.package_version == "2.0.1"
    assert job.variables == {
        "SENTRY_DSN": "http://sentry.dsn.somewhere",
        "SSH_USERNAME": "demo",
        "ENV": "dev",
        "SSH_SERVER": "dev.server.somewhere:22",
        "KAFKA_URL": "some.kafka.server:9999",
    }



from dotenv import load_dotenv

load_dotenv()

from getpass import getpass

import click

from kirby.demo import create_demo_db
from kirby.models import db
from kirby.models.security import user_datastore
from kirby.supervisor import run_supervisor
from kirby.web import app_maker
from smart_getenv import getenv

import logging

DEFAULT_LOG_FORMAT = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"


logging.basicConfig(
    level=logging.DEBUG,
    format=getenv("LOG_FORMAT", default=DEFAULT_LOG_FORMAT),
)
logging.getLogger("kafka").setLevel(logging.CRITICAL)

logger = logging.getLogger(__name__)


def read_topic(name):
    """
    debug function to consume the messages in a kafka topic
    :param name: name of the topic
    """
    from kafka import KafkaConsumer
    import msgpack
    import json
    from pprint import pformat

    bootstrap_servers = getenv(
        "KAFKA_BOOTSTRAP_SERVERS", type=list, separator=","
    )
    consumer = KafkaConsumer(
        name,
        bootstrap_servers=bootstrap_servers,
        auto_offset_reset="earliest",
        enable_auto_commit=True,
        value_deserializer=msgpack.loads,
    )

    for idx, message in enumerate(consumer, start=1):
        message = message.value
        try:
            message = json.loads(message.decode("utf-8"))
            logger.debug(pformat(message))
        except Exception:
            logger.debug(message)


@click.command()
@click.option(
    "--host", type=str, default="127.0.0.1", help="The interface to bind to"
)
@click.option("--port", type=str, default="8080", help="The port to bind to")
@click.option("--debug", type=bool, default=False, help="Start in DEBUG mode")
def web(host, port, debug):
    app = app_maker()
    app.run(debug=debug, port=port, host=host)


@click.command()
@click.argument("username")
def adduser(username):
    app = app_maker()

    with app.app_context():
        app.try_trigger_before_first_request_functions()
        user = user_datastore.find_user(username=username)
        if user is None:
            user = user_datastore.create_user(
                username=username, password=getpass()
            )

        msg = ""
        if click.confirm("Give admin rights?"):
            msg = "with admin rights"
            role = user_datastore.find_role("admin")
            user_datastore.add_role_to_user(user=user, role=role)

        click.echo(f"User {username} added {msg}")
        db.session.commit()


@click.command()
@click.argument("name")
@click.option(
    "--window",
    type=int,
    default=5,
    help="Leader election window size (in seconds)",
)
@click.option(
    "--wakeup",
    type=int,
    default=30,
    help="Shortest time interval between two scheduler executions",
)
def supervisor(name, window, wakeup):
    run_supervisor(name, window, wakeup)


@click.command()
def demo():
    app = app_maker()

    with app.app_context():
        app.try_trigger_before_first_request_functions()
        create_demo_db(db.session)

    click.echo("demo data inserted in the database")


@click.group()
def debug():
    pass


@click.command()
@click.argument("name")
def dump(name):
    read_topic(name)


debug.add_command(dump)


@click.group()
def cli():
    pass


cli.add_command(web)
cli.add_command(adduser)
cli.add_command(supervisor)
cli.add_command(demo)
cli.add_command(debug)


if __name__ == "__main__":
    cli()



from pytest import raises
from datetime import datetime

from dateutil.parser import parse
from kirby.models import (
    db,
    Job,
    Environment,
    JobType,
    Context,
    Schedule,
    Suspension,
    NotificationGroup,
    Script,
    Topic,
    ConfigKey,
    ConfigScope,
)


def test_it_creates_a_job(webapp):

    job = Job(
        name="retrieve cash register data",
        description="connects to the cash register API and fetches latest sale",
        type=JobType.SCHEDULED,
    )

    test_env = Environment(name="test")

    context = Context(environment=test_env)
    context.set_config(url="http://localhost:8000", loop=30, retry=True)

    schedule = Schedule(name="every two minutes", hour="*", minute="*/2")
    context.add_schedule(schedule)

    suspension = Suspension(start=parse("2019-01-01"), end=parse("2019-01-02"))
    schedule.add_suspension(suspension)

    group = NotificationGroup(name="admins")
    group.add_email("admin@local.local")

    job.add_context(context)
    job.add_notification(group, on_retry=True, on_failure=True)

    db.session.add(job)
    db.session.commit()

    assert (
        job.notifications[0].groups[0].emails[0].email == "admin@local.local"
    )


def test_it_creates_a_script(webapp):
    test_env = Environment(name="test_env")
    db.session.add(test_env)
    job = Job(
        name="retrieve cash register data",
        description="connects to the cash register API and fetches latest sale",
        type=JobType.SCHEDULED,
    )
    db.session.add(job)

    context = Context(environment=test_env, job=job)

    schedule = Schedule(name="every two minutes", hour="*", minute="*/2")
    db.session.add(schedule)
    context.add_schedule(schedule)
    db.session.add(context)

    script = Script(
        context=context,
        package_name="test_package",
        package_version="0.0.1",
        first_seen=datetime.now(),
    )

    source = Topic(name="source")
    db.session.add(source)
    destination = Topic(name="destination")
    db.session.add(destination)
    script.add_source(source)
    script.add_destination(destination)

    db.session.add(script)
    db.session.commit()

    assert script.sources == [source]
    assert script.destinations == [destination]
    assert script.context.environment.name == "test_env"
    assert script.context.job.name == "retrieve cash register data"
    assert script.context.schedules[0].name == "every two minutes"


def test_it_validate_schedule(webapp):
    with raises(ValueError) as excinfo:
        schedule = Schedule(name="Schedule test", hour="/2", minute="/3")
        db.session.add(schedule)
        db.session.commit()
    assert "hour" in str(
        excinfo.value
    ), f"The Schedule should not accept '/2' for hour"

    name_schedule_2 = "Schedule test II"

    schedule_2 = Schedule(name=name_schedule_2, hour="*/2")
    db.session.add(schedule_2)
    db.session.commit()

    schedule_2_commit = (
        db.session.query(Schedule).filter_by(name=name_schedule_2).one()
    )
    assert schedule_2_commit.hour == "*/2"

    with raises(ValueError):
        schedule_2.minute = "/3"
        db.session.commit()


def test_it_can_associate_config_to_context(webapp):
    test_env = Environment(name="test_env")
    db.session.add(test_env)
    job = Job(name="retrieve cash register data", type=JobType.SCHEDULED)
    job.set_config(API_URL="http://a.place.somewhere", OPERATION_MODE="normal")
    db.session.add(job)

    context = Context(environment=test_env, job=job)
    db.session.add(context)

    config = ConfigKey(name="API_URL", value="http://another.place.somewhere")
    config.context = context

    db.session.commit()

    assert config.scope == ConfigScope.CONTEXT

    assert context.variables() == {
        "API_URL": "http://another.place.somewhere",
        "OPERATION_MODE": "normal",
    }
